@@ -55,9 +55,6 @@ function MBTiles(uri, callback) {
     this._batchSize = +uri.query.batch;
     Step(function() {
         mbtiles._db = new sqlite3.Database(mbtiles.filename, this);
-    }, function(err) {
-        if (err) throw err;
-        mbtiles._setup(this);
     }, function(err) {
         if (err) throw err;
         fs.stat(mbtiles.filename, this);
@@ -79,6 +76,7 @@ MBTiles.registerProtocols = function(tilelive) {
 MBTiles.list = function(filepath, callback) {
     filepath = path.resolve(filepath);
     fs.readdir(filepath, function(err, files) {
+        if (err && err.code === 'ENOENT') return callback(null, {});
         if (err) return callback(err);
         for (var result = {}, i = 0; i < files.length; i++) {
             var name = files[i].match(/^([\w-]+)\.mbtiles$/);
@@ -174,7 +172,7 @@ MBTiles.prototype.getTile = function(z, x, y, callback) {
             } else {
                 var options = {
                     'Content-Type': MBTiles.utils.getMimeType(row.tile_data),
-                    'Last-Modified': mbtiles._stat.mtime,
+                    'Last-Modified': new Date(mbtiles._stat.mtime).toUTCString(),
                     'ETag': mbtiles._stat.size + '-' + Number(mbtiles._stat.mtime)
                 };
                 return callback(null, row.tile_data, options);
@@ -215,22 +213,27 @@ MBTiles.prototype.getGrid = function(z, x, y, callback) {
             }
             if (err) return callback(err);
 
-            try {
-                var grid = zlib.inflate(
-                           !Buffer.isBuffer(row.grid)
-                           ? new Buffer(row.grid, 'binary')
-                           : row.grid
-                       ).toString();
-                var data = rows.reduce(function(memo, r) {
-                        memo[r.key_name] = JSON.parse(r.key_json);
-                        return memo;
-                    }, {});
-                var result = _(JSON.parse(grid)).extend({ data: data });
-            } catch (err) {
-                return callback(new Error('Grid is invalid'));
-            }
-
-            callback(null, result);
+            zlib.inflate(!Buffer.isBuffer(row.grid)
+                ? new Buffer(row.grid, 'binary')
+                : row.grid,
+                function(err,buffer) {
+                    if (err) return callback(new Error('Grid is invalid:' + err.message));
+                    try {
+                        var data = rows.reduce(function(memo, r) {
+                            memo[r.key_name] = JSON.parse(r.key_json);
+                            return memo;
+                        }, {});
+                        var result = JSON.parse(buffer.toString());
+                    } catch(err) {
+                        return callback(new Error('Grid is invalid:' + err.message));
+                    }
+                    var options = {
+                        'Content-Type': 'text/javascript',
+                        'Last-Modified': new Date(that._stat.mtime).toUTCString(),
+                        'ETag': that._stat.size + '-' + Number(that._stat.mtime)
+                    };
+                    callback(null, _(result).extend({data:data}), options);
+                });
         }
     );
 };
@@ -250,6 +253,15 @@ MBTiles.prototype._metadata = function(key, callback) {
         });
 };
 
+
+MBTiles.prototype._close = function(callback) {
+    this._db.close(callback);
+};
+
+MBTiles.prototype.close = function(callback) {
+    this._close(callback);
+};
+
 MBTiles.prototype._getInfo = function(callback) {
     var mbtiles = this;
     var info = {};
@@ -385,18 +397,24 @@ MBTiles.prototype.startWriting = function(callback) {
     if (typeof callback !== 'function') throw new Error('Callback needed');
     if (!this.open) return callback(new Error('MBTiles not yet loaded'));
 
-    // Sets the synchronous flag to OFF for (much) faster inserts.
-    // See http://www.sqlite3.org/pragma.html#pragma_synchronous
-
     var mbtiles = this;
-    if (!this._isWritable) {
-        this._isWritable = 1;
-        this._clearCaches();
-        this._db.run('PRAGMA synchronous=OFF', callback);
-    } else {
-        this._isWritable++;
-        return callback(null);
-    }
+    Step(function() {
+        mbtiles._setup(this);
+    }, function(err) {
+        if (err) throw err;
+        // Sets the synchronous flag to OFF for (much) faster inserts.
+        // See http://www.sqlite3.org/pragma.html#pragma_synchronous
+        if (!mbtiles._isWritable) {
+            mbtiles._isWritable = 1;
+            mbtiles._clearCaches();
+            mbtiles._db.run('PRAGMA synchronous=OFF', this);
+        } else {
+            mbtiles._isWritable++;
+            this();
+        }
+    }, function(err) {
+        return callback(err);
+    });
 };
 
 MBTiles.prototype._clearCaches = function() {
@@ -529,7 +547,9 @@ MBTiles.prototype.putTile = function(z, x, y, data, callback) {
     // Flip Y coordinate because MBTiles files are TMS.
     y = (1 << z) - 1 - y;
 
-    var id = crypto.createHash('md5').update(data).digest('hex');
+    // Tilelive may pass us a data.key. If not, generate an md5
+    // from the image buffer data.
+    var id = data.key || crypto.createHash('md5').update(data).digest('hex');
     if (!this._tileCache[id]) {
         // This corresponds to the images table.
         this._tileCache[id] = data;
@@ -545,6 +565,31 @@ MBTiles.prototype.putTile = function(z, x, y, data, callback) {
     else return callback(null);
 };
 
+// Inserts a duplicate tile into the MBTiles store. Scheme is XYZ.
+//
+// - @param {Number} z tile z coordinate
+// - @param {Number} x tile x coordinate
+// - @param {Number} y tile y coordinate
+// - @param {Number} duplicate tile ID
+// - @param {Function(err)} callback
+MBTiles.prototype.putDuplicateTile = function(z, x, y, id, callback) {
+    if (typeof callback !== 'function') throw new Error('Callback needed');
+    if (!this.open) return callback(new Error('MBTiles not yet loaded'));
+    if (!this._isWritable) return callback(new Error('MBTiles not in write mode'));
+
+    // Flip Y coordinate because MBTiles files are TMS.
+    y = (1 << z) - 1 - y;
+
+    // This corresponds to the map table.
+    var coords = hash(z, x, y);
+    if (!this._mapCache[coords]) this._mapCache[coords] = { z: z, x: x, y: y };
+    this._mapCache[coords].tile_id = id;
+
+    // Only commit when we can insert at least batchSize rows.
+    if (++this._pending >= this._batchSize) return this._commit(callback);
+    else return callback(null);
+};
+
 // Inserts a grid into the MBTiles store. Scheme is XYZ.
 //
 // - @param {Number} z grid z coordinate
@@ -562,16 +607,22 @@ MBTiles.prototype.putGrid = function(z, x, y, data, callback) {
 
     // Preprocess grid data.
     var json = JSON.stringify({ grid: data.grid, keys: data.keys });
-    var id = crypto.createHash('md5').update(json).digest('hex');
+    var id = data.id || crypto.createHash('md5').update(json).digest('hex');
     if (!this._gridCache[id]) {
         // This corresponds to the grid_utfgrid table.
-        this._gridCache[id] = zlib.deflate(new Buffer(json, 'utf8'));
+        var that = this;
+        Step(function() {
+            zlib.deflate(new Buffer(json, 'utf8'),this);
+        }, function(err,buffer) {
+            if (err) return callback(err);
+            that._gridCache[id] = buffer;
+            // This corresponds to the grid_key table.
+            that._keyCache[id] = Object.keys(data.data || {});
 
-        // This corresponds to the grid_key table.
-        this._keyCache[id] = Object.keys(data.data || {});
-
-        // This corresponds to the keymap table.
-        _(this._dataCache).extend(data.data || {});
+            // This corresponds to the keymap table.
+            _(that._dataCache).extend(data.data || {});
+        }
+        );
     }
 
     // This corresponds to the map table.
@@ -609,3 +660,4 @@ MBTiles.prototype.putInfo = function(data, callback) {
         });
     });
 };
+
